/*====================================================================================
   This file is part of the TechTools Plugin Development Kit. 
   Copyright (c) 2011 by TechTools  

   You may modify this file and/or create derived works to develop plugins 
   for use with TechTools's DigiView software and capture files obtained from
   TechTools' hardware. Any other use is prohibited.

            email: support@tech-tools.com,  sales@tech-tools.com 
		web: www.tech-tools.com   Voice:972-272-9392  FAX: 972-494-5814
======================================================================================*/

#include "../plugin.h"

/* A full pugin to interpret a sepcific sync type device: DAC8045S45.
   Demonstrates edge detection, inversion, framing, idle fields, lookup tables
   and maintaining context through static vars and a state-machine
*/

// A little short-hand... These give readable names to the Configuration Options and
// field formats described below.  The final entries are used as sentinals to mark the length/ends  

//  configuration items (case 1)
static enum cfgnames {  CLK_MASK,DATA_MASK,EN_MASK,SHOWIDLE,SHOWPADS,CFGEND};
// field formats (case 2)
static enum fieldnames { F_CMD, F_DATA, F_TRUNC, F_PAD, FEND};

// protocol parser states
static enum states { ST_IDLE, ST_GETCMD, ST_GETDATA, ST_GETPAD};

// unused stubs
void OnLoad() {}
void SetInitItem(unsigned char ID, unsigned char subID, int value) {}
void OnUnLoad() {}
void EndOfData() {} // we do not need to do any clean up.


// called to retrieve descriptions and configuration information from the plug-in
// fills the provided vector container with the requested set of strings
void GetStrList(int ID, vector<string> &strl)
{
	// You are free to store your strings any way you wish as long as you
	// copy them into the provided vector when requested. You might choose
	// to store them as constant arrays or pre-initialized vectors.  We chose
	//to load them directly in-line here, keeping all configuration in one place.
	switch (ID)
	{ 
	case 0:   // Plug-in Description
		// These lines are displayed in the signal editor for identification
		// but serve no other purpose. For best results, keep to 4 or fewer lines.
		strl.push_back(	"Full DAC8045S085 Example V0.1 BETA");
		break;

	case 1:   // configuration options
		// These are the items displayed in the SIGNAL EDITOR. They allow the
		// user to configure the plug-in for a given signal. You must provide
		// at least 1 chanselect type option to allow the user to assign physical
		// channel(s) to the signal. 
		strl.assign(CFGEND,"");  // allocate and init the list
		strl[CLK_MASK]     =  "Clock Channel,chanselect,1,1,1,true,false";
		strl[DATA_MASK]    =  "Data Channel,chanselect,2,1,1,true,false";
		strl[EN_MASK]      =  "/Enable Channel,chanselect,4,1,1,true,false";
		strl[SHOWIDLE]     =  "Show Idles,checkbox,false";
		strl[SHOWPADS]     =  "Show Pad Fields,checkbox,false";
		break;

	case 2:  // Field Formats
		// These tell the app how to display your fields. You must define at least 
		// one field format.
		strl.assign(FEND,"");  // allocate and init the list
		strl[F_CMD]   =	"CMD,BLUE,WHITE,{0%3:0}";
		strl[F_DATA]  =	"DATA,c0c0c0,BLACK,{7:0}";
		strl[F_TRUNC] =	"TRUNC,80FFFF,BLACK,T:({7:0})";
		strl[F_PAD]   = "PAD,c0c0c0,BLACK,P:{3:0}";
		break;

	case 3:  //return which preparser we need
		strl.push_back("RAW");
		break;

	case 4: // return minimim framework ver we need
		strl.push_back("1");
		break;

	case 5: //lookup strings
		strl.push_back("0,0,WT A");
		strl.push_back("0,1,WT B");
		strl.push_back("0,2,WT C");
		strl.push_back("0,3,WT D");
		strl.push_back("0,4,Update A");
		strl.push_back("0,5,Update B");
		strl.push_back("0,6,Update C");
		strl.push_back("0,7,Update D");
		strl.push_back("0,8,Update All");
		strl.push_back("0,9,PwrDwn");    
	}
}


static bool first_sample;

// our configuration items
static Data64 datamask;
static bool InvertDataFlag;
static Data64 clkmask;
static bool InvertClkFlag;
static Data64 enmask;
static bool InvertEnFlag;
static bool showidle;
static bool showpads;

// called to set our configuration options.
// These represent the choices the user made in response to your Configuration Options
// You can store and retrieve your configuration items any way you want.
// We chose to store them directly in named variables.
// The ID corresponds to the index of the Configuration Options item.
// eg 0 => the first configuration string you provided....

void SetCfgItem(unsigned char ID, unsigned char subID, int value)
{
	switch (ID)
	{
	case DATA_MASK:  // data bus mask
		switch (subID)
		{
		case 0: InvertDataFlag = (value & 0x01); break;
		case 1: datamask.lowint = value;   break;
		case 2: datamask.highint = value;  break;
		}
		break;

	case CLK_MASK: 
		switch (subID)
		{
		case 0: InvertClkFlag = (value & 0x01); break;
		case 1: clkmask.lowint = value;  break;
		case 2: clkmask.highint = value; break;
		}
		break;

	case EN_MASK: 
		switch (subID)
		{
		case 0: InvertEnFlag = (value & 0x01); break;
		case 1: enmask.lowint = value;  break;
		case 2: enmask.highint = value; break;
		}
		break;
	case SHOWIDLE: showidle = (value != 0); break;
	case SHOWPADS: showpads = (value != 0); break;
	// ignore illegal cfg items
	}
}


// Called immediately after setting configuration options and before data
// streaming begins. Do any initialization you require here. In particular,
// note that you should assume that the configuration options have changed.
void StartOfData()
{	
	// some of our initialization is dependent on SEEing the first sample...
	// do it in the first call to 'parse'
	first_sample = true;
}


// Called when a new data sample has been received.
// This is the protocol parser itself.  It must examine each data sample as it
// arrives, send back field information when appropriate for this protocol and
// return immediately.  It must never loop or attempt to use stdio.  It must 
// maintain its state between calls through statics.
// NOTE that this could get called on the order of 250,000 times per capture per signal,
// so efficiency is important.  We pre-converted many of the configurtion items to 
// bools in the setcfg routine to minimize masking, shifting and int64 compares here.
void Parse(int64 timestamp, Data64 rawdata)
{
    static int shifter;        // stores gathered data bits
	static int bitcount;       // tracks # bits gathered
	static bool wasclkbit;     // remembers clk level for edge detection
	static bool wasenablebit;  // remembers enabled state for edge detection
	static int state;          // current protocol statef
	static int64 starttime;    // timestamp of the first bit of the current field
	static int bitswanted;     // bits need to fill the current field

	bool clkbit;
    bool enablebit;
    bool databit;
	Data64 mydata;

	// extract our control signals.
	// NOTE: the PACK operation is expensive (even with pre-calculated limits).  It should only
	// be used with bus type signals (> 1 bit) and only when needed.  For single bit signals, 
	// it is far more efficient to just mask directly and convert to a bool (avoids int64 compares later).
	clkbit    = ((rawdata.i64 & clkmask.i64)  == 0);  // falling edge is active edge
	enablebit = ((rawdata.i64 & enmask.i64)   == 0);  // active low
	databit   = ((rawdata.i64 & datamask.i64) != 0);

	if (InvertClkFlag)  clkbit = !clkbit;
	if (InvertEnFlag)   enablebit = !enablebit;
	if (InvertDataFlag) databit = !databit;

	// if this is the first sample, establish a baseline by noting the current
	// control signal levels and then return.
	if (first_sample)
	{ 
		first_sample = false;  // only one 'first sample'
		bitcount = 0;
		shifter = 0;
		state = ST_IDLE;

		wasclkbit = clkbit;           // note current clock level
		wasenablebit   = enablebit;   //   and enabled state
		return;
	}

	if (!enablebit & wasenablebit)  // disable edge
	{
		if (bitcount > 0) // partial data...send it
		{
			if (bitswanted > bitcount) 
				shifter <<= (bitswanted-bitcount);  // top justified the partial data 
			mydata.i64 = 0;
			mydata.bytes[0] = shifter;
			if (state == ST_GETCMD)
			   StartFrame(starttime,mydata,F_TRUNC);
			else
				StartField(starttime,mydata,F_TRUNC);

			EndFrame(timestamp);
		}

		state = ST_GETCMD;  // gather the command bits
		shifter = 0;
		bitcount = 0;
		bitswanted = 4;
	}

	if (enablebit && !wasclkbit && clkbit)  // enabled and proper clock edge
	{
		if (bitcount == 0) starttime = timestamp;  // remember first bit time
		shifter <<= 1;
		if (databit) shifter++;
		bitcount++;
		if (bitcount >= bitswanted)
		{ 
			switch (state)
			{
			case ST_IDLE:
				break;  // do nothing until enable edge

			case ST_GETCMD:
				mydata.i64 = 0;
				// adjust the command data to select the string we want
				switch (shifter & 0x03) 
				{ 
				case 0: // wt, no update
					mydata.bytes[0] = shifter >> 2;   // first 4 lookup strings
					break;
				case 1: // wt with update
					mydata.bytes[0] = shifter & 0x0C; // next 4 lookup strings
					break;
				case 2: 
					mydata.bytes[0] = 8;  // update all
					break;
				case 3: 
					mydata.bytes[0] = 9;  // power down
					break;
				}

				// send command
				StartFrame(starttime,mydata,F_CMD);
				if (showidle) EndField(timestamp);
				state = ST_GETDATA;
				bitswanted = 8;
				break;

			case ST_GETDATA:
				// send data
				mydata.i64 = 0;
				mydata.bytes[0] = shifter;
				StartField(starttime,mydata,F_DATA);
				if (showidle) EndField(timestamp);
				state = ST_GETPAD;
				bitswanted = 4;
				break;

			case ST_GETPAD:
				mydata.i64 = 0;
				mydata.bytes[0] = shifter;
				if (showpads) StartField(starttime,mydata,F_PAD);
				EndFrame(timestamp);
				state = ST_IDLE; 	// spec says ignore extra clocks...idle until new select
				bitswanted = 0;
				break;
			}

			shifter = 0;
			bitcount = 0;
		}
	}
	wasclkbit = clkbit;           // note current clock level
	wasenablebit   = enablebit;   // and enabled state

}  // parser()
